# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.bulk_data
:Synopsis:          This module includes functions that relate to the Bulk Data API.
:Usage:             ``from khoros import bulk_data``
:Example:           ``base_url = bulk_data.get_base_url(community_id='example.prod')``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     14 Mar 2022
"""

import requests

from . import errors
from .utils import log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)


def get_base_url(khoros_object=None, community_id=None, europe=False):
    """This function constructs and/or retrieves the base URL for the Bulk Data API.

    .. versionadded:: 5.0.0

    .. note:: The URL from the helper settings will be leveraged when available unless the ``community_id`` is
              explicitly defined as a function parameter.

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros], None
    :param community_id: The Community ID to leverage in the URL
    :type community_id: str, None
    :param europe: Determines if the European URL should be utilized (``False`` by default)
    :type europe: bool
    :returns: The base URL for the Bulk Data API
    """
    base_url = None
    base_urls = {
        True: 'https://eu.api.lithium.com/lsi-data/v2/data/export/community/',
        False: 'https://api.lithium.com/lsi-data/v2/data/export/community/'
    }

    # Retrieve the base URL from the helper settings when defined
    if not community_id and khoros_object:
        try:
            base_url = khoros_object.bulk_data_settings['base_url']
        except (KeyError, AttributeError):
            # Attempt to define the community ID if found
            try:
                community_id = khoros_object.bulk_data_settings['community_id']
            except (KeyError, AttributeError):
                pass

    # Construct the base URL when not defined in the helper settings
    if not base_url:
        europe = False if europe is None else europe
        base_url = base_urls.get(europe)
        if community_id:
            base_url = f'{base_url}{community_id}'
    return base_url


def query(khoros_object=None, community_id=None, client_id=None, token=None, from_date=None, to_date=None, fields=None,
          europe=None, export_type=None, full_response=False):
    """This function performs a query against the Bulk Data API to retrieve CSV or JSON data.

    .. versionadded:: 5.0.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros], None
    :param community_id: The Community ID to leverage in the URL
    :type community_id: str, None
    :param client_id: The Client ID used to authenticate to the Bulk Data API
    :type client_id: str, None
    :param token: The access token used to authenticate to the Bulk Data API
    :type token: str, None
    :param from_date: The **From** Date in ``YYYYmmDD`` or ``YYYYmmDDhhMM`` format.
    :type from_date: str, None
    :param to_date: The **To** Date in ``YYYYmmDD`` or ``YYYYmmDDhhMM`` format.
    :type to_date: str, None
    :param fields: Optional fields to include in the data export as a comma-separated string or iterable
    :type fields: str, list, tuple, set, None
    :param europe: Determines if the European URL should be utilized (``False`` by default)
    :type europe: bool
    :param export_type: Determines if the data should be returned in ``csv`` (default) or ``json`` format
    :type export_type: str, None
    :param full_response: Determines if the full :py:mod:`requests` object should be returned (``False`` by default)
    :type full_response: bool
    :returns: The CSV or JSON data for the Bulk Data API request (or the full :py:mod:`requests` object)
    :raises: :py:exc:`TypeError`, :py:exc:`ValueError`,
             :py:exc:`khoros.errors.exceptions.MissingAuthDataError`,
             :py:exc:`khoros.errors.exceptions.APIRequestError`
    """
    # Get the base URL
    if khoros_object and khoros_object.bulk_data_settings.get('base_url'):
        base_url = khoros_object.bulk_data_settings.get('base_url')
    else:
        base_url = get_base_url(khoros_object, community_id, europe)

    # Get the client ID
    if not client_id:
        if khoros_object and khoros_object.bulk_data_settings.get('client_id'):
            client_id = khoros_object.bulk_data_settings.get('client_id')
        else:
            raise errors.exceptions.MissingAuthDataError('A valid Client ID is required to utilize the Bulk Data API.')

    # Get the auth token
    if not token:
        if khoros_object and khoros_object.bulk_data_settings.get('token'):
            token = khoros_object.bulk_data_settings.get('token')
        else:
            raise errors.exceptions.MissingAuthDataError('A valid access token is required to utilize the '
                                                         'Bulk Data API.')

    # Construct the API headers
    headers = _construct_headers(khoros_object, client_id, export_type)

    # Construct the authentication tuple
    auth = (token, '')

    # Construct the parameters
    params = _construct_parameters(from_date, to_date, fields)

    # Perform the API call
    response = requests.get(base_url, params=params, auth=auth, headers=headers)
    if not full_response:
        if response.status_code != 200:
            raise errors.exceptions.APIRequestError(f'Bulk Data API request failed with a {response.status_code} '
                                                    f'response.')
        if export_type.lower() == 'json':
            response = response.json()
        else:
            response = response.text
    return response


def _construct_parameters(_from_date=None, _to_date=None, _fields=None):
    """This function constructs the parameters to utilize in the API call.

    .. versionadded:: 5.0.0

    :param _from_date: The beginning date range parameter
    :type _from_date: str, None
    :param _to_date: The ending date range parameter
    :type _to_date: str, None
    :param _fields: The optional list of fields to include
    :type _fields: str, list, tuple, set, None
    :returns: The dictionary of parameters to use in the API call
    :raises: :py:exc:`TypeError`
    """
    # Validate and add the date parameters
    _validate_date_field(_from_date)
    _validate_date_field(_to_date)
    _params = {
        'fromDate': _from_date,
        'toDate': _to_date,
    }

    # Validate and add the fields parameter if applicable
    if _fields:
        if not isinstance(_fields, str):
            if isinstance(_fields, list) or isinstance(_fields, tuple) or isinstance(_fields, set):
                _fields = ','.join(_fields)
            else:
                raise TypeError('The fields parameter should be a comma-separated string or an iterable.')
        _params['fields'] = _fields
    return _params


def _validate_date_field(_date_value):
    """This function validates the ``fromDate`` and ``toDate`` fields to ensure they are in a valid format.

    .. versionadded:: 5.0.0

    :param _date_value: The date value to be evaluated (e.g. ``20220313``)
    :type _date_value: str
    :returns: None
    :raises: :py:exc:`ValueError`
    """
    if not isinstance(_date_value, str) or not _date_value.isnumeric() or \
            (len(_date_value) != 8 and len(_date_value) != 12):
        raise ValueError('The fromDate and toDate fields should be in yyyyMMdd or yyyyMMddHHmm format.')


def _construct_headers(_khoros_object=None, _client_id=None, _export_type=None):
    """This function constructs the headers to use in a Bulk Data API call.

    .. versionadded:: 5.0.0

    :param _khoros_object: The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros], None
    :param _client_id: The Client ID to use when authenticating the API calls
    :type _client_id: str, None
    :param _export_type: Indicates the export type as either ``csv`` (default) or ``json``
    :type _export_type: str
    :returns: A dictionary containing the API headers
    :raises: :py:exc:`khoros.errors.exceptions.MissingAuthDataError`
    """
    # Get the client ID
    if not _client_id:
        if _khoros_object and _khoros_object.bulk_data_settings.get('client_id'):
            _client_id = _khoros_object.bulk_data_settings.get('client_id')
        else:
            raise errors.exceptions.MissingAuthDataError('A valid Client ID is required to utilize the Bulk Data API.')

    # Get the Accept value depending on the export type
    if not _export_type:
        if _khoros_object and _khoros_object.bulk_data_settings.get('export_type'):
            _export_type = _khoros_object.bulk_data_settings.get('export_type')
        else:
            # Default to CSV export
            _export_type = 'csv'
    _accept_value = _get_export_header(_export_type)

    # Construct and return the header
    _headers = {
        'client-id': _client_id,
        'Accept': _accept_value
    }
    return _headers


def _get_export_header(_export_type='csv'):
    """This function retrieves the appropriate ``Accept`` header value depending on the export type.

    .. versionadded:: 5.0.0

    :param _export_type: Indicates the export type as either ``csv`` (default) or ``json``
    :type _export_type: str
    :returns: The appropriate ``Accept`` header value
    """
    if _export_type.lower() == 'json':
        _export_header_value = 'application/json'
    else:
        # TODO: Add warning log entry if not CSV
        _export_header_value = 'text/csv'
    return _export_header_value



# -*- coding: utf-8 -*-
"""
:Package:        khoros.errors
:Synopsis:       This module includes custom exceptions
:Usage:          ``import khoros.errors`` (Imported by default in primary package)
:Example:        ``raise errors.exceptions.MissingAuthDataError``
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  01 Jul 2020
"""
# Define all modules that will be imported with the "import *" method
__all__ = ['exceptions', 'handlers', 'translations']

# Always import the warnings package and the exceptions module
from . import exceptions, handlers, translations



# -*- coding: utf-8 -*-
"""
:Module:            khoros.liql
:Synopsis:          Collection of functions and classes to leverage the Community API v2 and LiQL for searches
:Usage:             ``from khoros import liql`` (Imported by default in :py:mod:`khoros.core` package)
:Example:           ``query_url = liql.format_query("SELECT * FROM messages WHERE id = '2' LIMIT 1")``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     23 May 2022
"""

from . import api, errors
from .utils import log_utils
from .utils.core_utils import convert_set

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)

COLLECTIONS = ['albums', 'attachments', 'boards', 'bookmarks', 'categories', 'communities', 'custom_tags',
               'floated_messages', 'grouphubs', 'images', 'inbox_notes', 'kudos', 'labels', 'me_toos',
               'membership_requests', 'messages', 'metrics', 'nodes', 'notes_threads', 'notification_feeds',
               'outbox_notes', 'product_categories', 'products', 'ranks', 'ratings', 'review_comments',
               'review_dimensions', 'review_ratings', 'reviews', 'roles', 'subscriptions', 'tags', 'threaded_notes',
               'tkb_helpfulness_ratings', 'users', 'videos']


def format_query(query, pretty_print=False, track_in_lsi=False, always_ok=False, error_code='', format_statements=True):
    """This function formats and URL-encodes a raw LiQL query to be able to use it within a Community v2 API URL.

    .. versionchanged:: 2.1.0
       Queries ending in a semicolon (``;``) will have that character stripped to avoid syntax errors.

    .. versionchanged:: 2.0.0
       Added URL-encoding support for several additional characters.

    :param query: The LiQL query to be formatted and url-encoded
    :type query: str
    :param pretty_print: Defines if the response should be "pretty printed" (``False`` by default)
    :type pretty_print: bool
    :param track_in_lsi: Defines if the query should be tracked within LSI (``False`` by default)
    :type track_in_lsi: bool
    :param always_ok: Defines if the HTTP response should **always** be ``200 OK`` (``False`` by default)
    :type always_ok: bool
    :param error_code: Allows an error code to optionally be supplied for testing purposes (ignored by default)
    :type error_code: str
    :param format_statements: Determines if statements (e.g. ``SELECT``, ``FROM``, et.) should be formatted to be in
                              all caps (``True`` by default)
    :type format_statements: bool
    :returns: The properly formatted query to be inserted in the URL
    """
    chars_to_encode = {
        ' ': '+',
        '=': '%3D',
        '"': '%22',
        '\'': '%27',
        '(': '%28',
        '}': '%29',
        '@': '%40'
    }
    query_statements = {
        'select': 'SELECT',
        'from': 'FROM',
        'where': 'WHERE',
        'order by': 'ORDER BY',
        'desc': 'DESC ',
        'limit': 'LIMIT',
        'asc': 'ASC',
        'offset': 'OFFSET'
    }
    parameters = {
        'pretty_print': '&api.pretty_print=true',
        'track_in_lsi': '&api.for_ui_search=true',
        'always_ok': '&api.always_ok',
        'error_code': '&api.error_code='
    }
    replacements = [chars_to_encode]
    if format_statements:
        replacements.append(query_statements)
    for criteria in replacements:
        for orig_string, new_string in criteria.items():
            query = query.replace(orig_string, new_string)
            if query.endswith(';'):
                query = query[:-1]
    if pretty_print:
        query = f"{query}{parameters.get('pretty_print')}"
    if track_in_lsi:
        query = f"{query}{parameters.get('track_in_lsi')}"
    if always_ok:
        query = f"{query}{parameters.get('always_ok')}"
    if error_code != '':
        query = f"{query}{parameters.get('error_code')}{error_code}"
    return query


def get_query_url(core_dict, query, pretty_print=False, track_in_lsi=False, always_ok=False,
                  error_code='', format_statements=True):
    """This function defines the full Community API v2 query URL for the LiQL query.

    :param core_dict: The ``core`` dictionary defined in the :py:class:`khoros.core.Khoros` object
    :type core_dict: dict
    :param query: The LiQL query to be encoded and embedded in the URL
    :type query: str
    :param pretty_print: Defines if the response should be "pretty printed" (``False`` by default)
    :type pretty_print: bool
    :param track_in_lsi: Defines if the query should be tracked within LSI (``False`` by default)
    :type track_in_lsi: bool
    :param always_ok: Defines if the HTTP response should **always** be ``200 OK`` (``False`` by default)
    :type always_ok: bool
    :param error_code: Allows an error code to optionally be supplied for testing purposes (ignored by default)
    :type error_code: str
    :param format_statements: Determines if statements (e.g. ``SELECT``, ``FROM``, et.) should be formatted to be in
                              all caps (``True`` by default)
    :type format_statements: bool
    :returns: The full Community API v2 URL in string format
    """
    query = format_query(query, pretty_print, track_in_lsi, always_ok, error_code, format_statements)
    return f"{core_dict['v2_base']}/search?q={query}"


def perform_query(khoros_object, query_url=None, liql_query=None, return_json=True, verify_success=False,
                  allow_exceptions=True, verify=None, return_items=False):
    """This function performs a LiQL query using full Community API v2 URL containing the query."

    .. versionchanged:: 5.0.0
       Two ``if`` statements have been merged.

    .. versionchanged:: 4.1.0
       The JSON response can now be reduced to just the returned items by passing ``return_items=True``.

    .. versionchanged:: 3.4.0
       Support has been introduced for the ``verify`` parameter to determine if SSL certificate verification is needed.

    .. versionchanged:: 2.3.0
       Added the ``allow_exceptions`` argument (``True`` by default) to allow exceptions to be disabled
       and updated the error/exception message to be more specific.

    .. versionchanged:: 2.0.0
       Added the ability for a raw LiQL query to be passed to this function rather than only a formatted query URL,
       and included the optional ``verify_success`` Boolean argument to verify a successful response.

    :param khoros_object: The Khoros object initialized via the :py:mod:`khoros.core` module
    :type khoros_object: class[khoros.Khoros]
    :param query_url: The full Khoros Community API v2 URL for the query
    :type query_url: str, None
    :param liql_query: The LiQL query to be performed, not yet embedded in the query URL
    :type liql_query: str, None
    :param return_json: Determines if the response should be returned in JSON format (``True`` by default)
    :type return_json: bool
    :param verify_success: Optionally check to confirm that the API query was successful (``False`` by default)
    :type verify_success: bool
    :param allow_exceptions: Defines whether or not exceptions can be raised for responses returning errors

                             .. caution:: This does not apply to exceptions for missing required data.

    :type allow_exceptions: bool
    :param verify: Determines whether or not to verify the server's TLS certificate (``True`` by default)
    :type verify: bool, None
    :param return_items: Reduces the JSON response to be only the list of items returned from the LiQL response
                         (``False`` by default)

                         .. note:: If an error response is returned then an empty list will be returned.

    :type return_items: bool
    :returns: The API response (in JSON format by default)
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.MissingAuthDataError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    # Determine if TLS certificates should be verified during API calls
    verify = api.should_verify_tls(khoros_object) if verify is None else verify

    # Validate the LiQL query
    if not query_url and not liql_query:
        raise errors.exceptions.MissingRequiredDataError("An API query URL or a raw LiQL query must be provided.")
    if liql_query:
        query_url = get_query_url(khoros_object.core, liql_query)
    if 'header' not in khoros_object.auth:
        error_msg = "Cannot perform the query as an authorization header is not configured."
        raise errors.exceptions.MissingAuthDataError(error_msg)

    # Perform the API call and validate the data
    response = api.get_request_with_retries(query_url, return_json, auth_dict=khoros_object.auth, verify=verify)
    if verify_success and not api.query_successful(response):
        error_msg = errors.handlers.get_error_from_json(response, include_error_bool=False)[2]
        if allow_exceptions:
            raise errors.exceptions.GETRequestError(error_msg)
        errors.handlers.eprint(error_msg)
    if return_json:
        # Convert the response to JSON as needed
        if not isinstance(response, dict):
            response = response.json()

        # Reduce teh scope to just the returned items when requested
        if return_items:
            if response.get('data') and response.get('data').get('items'):
                response = response.get('data').get('items')
            else:
                # Return an empty list as no items were found
                response = []
                # TODO: Log an entry stating that the LiQL query failed
    return response


def get_total_count(khoros_object, collection, where_filter="", verify_success=True):
    """This function retrieves the total asset count from a given collection (e.g. ``categories``).

    :param khoros_object: The Khoros object initialized via the :py:mod:`khoros.core` module
    :type khoros_object: class[khoros.Khoros]
    :param collection: The collection object to use in the FROM clause of the LiQL query (e.g. ``users``)
    :type collection: str
    :param where_filter: An optional filter to use as the WHERE clause in the LiQL query
    :type where_filter: str
    :param verify_success: Determines if the API query should be verified as successful (``True`` by default)
    :type verify_success: bool
    :returns: The total count as an integer
    :raises: :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    query_syntax = parse_query_elements('count(*)', collection, where_filter)
    response = perform_query(khoros_object, liql_query=query_syntax, verify_success=verify_success)
    return response['data']['count']


def get_returned_items(liql_response, only_first=False):
    """This function prunes a full LiQL API response down to only the returned item(s).

    .. versionchanged:: 3.3.2
       The error handling has been improved to avoid :py:exc:`IndexError` exceptions from being raised when no items
       were found in the LiQL response.

    .. versionadded:: 3.2.0

    :param liql_response: The full JSON response from the LiQL query as a dictionary
    :type liql_response: dict
    :param only_first: Returns only the first item found rather than a list of items (``False`` by default)
    :type only_first: bool
    :returns: A list of items (or a single item) found within the LiQL response
    :raises: :py:exc:`khoros.errors.exceptions.LiQLParseError`
    """
    if 'status' not in liql_response:
        raise errors.exceptions.LiQLParseError()
    elif liql_response.get('status') != 'success':
        raise errors.exceptions.LiQLParseError(message=liql_response.get('message'))
    liql_items = liql_response['data']['items']
    if only_first:
        fail_msg = "No items were found in the LiQL response and therefore a NoneType value will be returned."
        try:
            if len(liql_items) > 0:
                liql_items = liql_items[0]
            else:
                logger.error(fail_msg)
                liql_items = None
        except IndexError:
            logger.error(fail_msg)
            liql_items = None
    return liql_items


def parse_query_elements(select_fields, from_source, where_filter="", order_by=None, order_desc=True, limit=0):
    """This function parses query elements to construct a full LiQL query in the appropriate syntax.

    :param select_fields: One or more fields to be selected within the SELECT statement (e.g. ``id``)
    :type select_fields: str, tuple, list, set
    :param from_source: The source of the data to use in the FROM statement (e.g. ``messages``)
    :type from_source: str
    :param where_filter: The filters (if any) to use in the WHERE clause (e.g. ``id = '2'``)
    :type where_filter: str, tuple, list, dict, set
    :param order_by: The field(s) by which to order the response data (optional)
    :type order_by: str, tuple, set, dict, list
    :param order_desc: Defines if the ORDER BY directionality is DESC (default) or ASC
    :type order_desc: bool
    :param limit: Allows an optional limit to be placed on the response items (ignored by default)
    :type limit: int
    :returns: The query response in JSON format
    :raises: :py:exc:`khoros.errors.exceptions.OperatorMismatchError`,
             :py:exc:`khoros.errors.exceptions.InvalidOperatorError`
    """
    # Properly format the provided SELECT fields
    select_fields = parse_select_fields(select_fields)

    # Establish the base syntax to return
    full_syntax = f"SELECT {select_fields} FROM {from_source}"

    # Append the WHERE clause to the syntax if provided
    if type(where_filter) != str:
        where_filter = parse_where_clause(where_filter)
    if where_filter != "":
        full_syntax = f"{full_syntax} WHERE {where_filter}"

    # Append the ORDER BY clause to the syntax is provided
    if order_by:
        order_direction = {True: 'DESC', False: 'ASC'}
        if type(order_by) in LiQLSyntax.container_types:
            order_by = convert_set(order_by)
            order_by = ','.join(order_by)
        order_by_clause = f"ORDER BY {order_by} {order_direction.get(order_desc)}"
        full_syntax = f"{full_syntax} {order_by_clause}"

    # Append the LIMIT clause to the syntax if provided
    limit_set = True if limit > 0 else False
    if limit_set:
        limit_syntax = {True: f"LIMIT {limit}", False: ""}
        full_syntax = f"{full_syntax} {limit_syntax.get(limit_set)}"

    # Return the fully parsed query syntax
    return full_syntax


def parse_select_fields(_select_fields):
    """This function parses the fields to be used in the SELECT statement of the LiQL query.

    .. versionchanged:: 3.5.0
       Refactored the function to leverage the :py:func:`isinstance` built-in function.

    .. versionchanged:: 3.4.0
       Renamed the function to adhere to PEP8 guidelines.

    :param _select_fields: The field(s) to be used in the SELECT statement
    :type _select_fields: str, tuple, list, set
    :returns: The properly formatted SELECT fields in string format (comma-separated)
    :raises: :py:exc:`khoros.errors.exceptions.InvalidFieldError`
    """
    if not any((isinstance(_select_fields, list), isinstance(_select_fields, str), isinstance(_select_fields, tuple),
                isinstance(_select_fields, set))):
        raise errors.exceptions.InvalidFieldError("Fields must be in string or iterable format when explicitly defined")
    if any((isinstance(_select_fields, tuple), isinstance(_select_fields, list), isinstance(_select_fields, set))):
        _select_fields = convert_set(_select_fields)
        _select_fields = ",".join(_select_fields)
    elif isinstance(_select_fields, str):
        _select_fields = _select_fields.replace(';', ',').replace(', ', ',')
    return _select_fields


def structure_cursor_clause(cursor=None, liql_response=None):
    """This function structures the CURSOR clause for a LiQL query.

    .. versionchanged:: 4.0.0
       The cursor string has been wrapped in single quotes to prevent the
       :py:exc:`khoros.errors.exceptions.LiQLParseError` exception from being raised.

    .. versionadded:: 3.5.0

    :param cursor: A cursor value from the ``next_cursor`` key in a LiQL response
    :type cursor: str, None
    :param liql_response: A full LiQL query response in dictionary format
    :type liql_response: dict, None
    :returns: The properly formatted CURSOR clause for use in a new LiQL query
    :raises: :py:exc:`khoros.errors.exceptions.InvalidFieldError`,
             :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    if not cursor and not liql_response:
        raise errors.exceptions.MissingRequiredDataError('Must provide cursor value or LiQL response')
    if cursor:
        if not isinstance(cursor, str):
            raise errors.exceptions.InvalidFieldError('Cursor value must be a string')
        if not cursor.lower().startswith('cursor'):
            cursor = f"CURSOR '{cursor}'"
    else:
        cursor = ''
        if liql_response.get('data') and liql_response['data'].get('next_cursor'):
            cursor = f"CURSOR {liql_response['data'].get('next_cursor')}"
    return cursor


def _wrap_string_values(_where_value):
    """This function wraps values going in the WHERE clause in single-quotes if they are not integers.

    .. versionchanged:: 3.4.0
       Renamed the function to adhere to PEP8 guidelines.

    :param _where_value: The value to be evaluated and potentially wrapped in single-quotes
    :returns: The value in int or string format
    """
    try:
        _where_value = int(_where_value)
    except (TypeError, ValueError):
        _where_value = f"'{_where_value}'"
    return _where_value


def _convert_where_dicts_to_lists(_dict_list):
    """This function converts dictionaries supplied as WHERE clause filters into properly formatted lists.

    .. versionchanged:: 3.4.0
       Renamed the function to adhere to PEP8 guidelines.

    :param _dict_list: A list of dictionaries with the WHERE clause information
    :type _dict_list: list
    :returns: A list of lists containing the WHERE clause information
    """
    _master_list = []
    for _where_dict in _dict_list:
        for _dict_key, _dict_val in _where_dict.items():
            _single_list = [_dict_key]
            if type(_dict_val) == tuple or type(_dict_val) == list or type(_dict_val) == set:
                _dict_val = convert_set(_dict_val)
                if len(_dict_val) == 2:
                    for _val in _dict_val:
                        _single_list.append(_val)
                else:
                    raise errors.exceptions.OperatorMismatchError
            else:
                _single_list.append(_dict_val)
            _master_list.append(_single_list)
    return _master_list


def parse_where_clause(where, join_logic='AND'):
    """This function parses the data supplied for the WHERE clause of a LiQL query.

    .. versionchanged:: 4.3.0
       Refactored the function to be more efficient and Pythonic, and added missing parenthesis on the
       exception classes.

    .. versionchanged:: 3.4.0
       Renamed the function to adhere to PEP8 guidelines and converted from a private to a public function.

    :param where: The WHERE clause information
    :type where: str, tuple, list, set, dict
    :param join_logic: The logic to use
    :type join_logic: str, tuple, list
    :returns: A properly formatted WHERE clause (excluding the WHERE statement at the beginning)
    :raises: :py:exc:`khoros.errors.exceptions.InvalidOperatorError`,
             :py:exc:`khoros.errors.exceptions.OperatorMismatchError`
    """
    # Examples:
    #   _where = ('id', 5)      #2-length tuple
    #   _where = ('id', '>', 5)     # 3-length tuple
    #   _where = (('id', 5), ('id', 6))     # tuple of tuples
    #   _where = [('replies.count(*)', '>', 5), ('replies.count(*)', '<', 10)]      # list of tuples
    #   _where = {('id', 5), ('replies.count(*)', '=', 0)}      # set
    #   _where = {'id': 5, 'replies.count(*)': 0}       # one-to-one dict
    #   _where = {'id': ('>', 5), 'id': ('<', 10)}      # one-to-two dict

    # Add them into a list as needed
    if not isinstance(where, list):
        if isinstance(where, dict):
            where = _convert_where_dicts_to_lists([where])
        if isinstance(where, tuple) or isinstance(where, set) and where:
            where = convert_set(where)
            if type(where[0]) not in LiQLSyntax.container_types:
                where = [where]

    # Determine the multi-clause logic to use
    # TODO: Figure out how to allow where clause grouping with logic
    if isinstance(join_logic, str):
        if join_logic not in LiQLSyntax.logic_operators:
            raise errors.exceptions.InvalidOperatorError()
        join_logic = [join_logic]
    elif len(join_logic) == 1:
        if join_logic[0] not in LiQLSyntax.logic_operators:
            raise errors.exceptions.InvalidOperatorError()
        else:
            join_logic = [join_logic[0]]
    else:
        if len(join_logic) != len(where) - 1:
            raise errors.exceptions.OperatorMismatchError()
        if not isinstance(join_logic, list):
            join_logic = list(join_logic)

    # Parse the where clause
    full_clause = ""
    num_clauses = len(where)
    while num_clauses > 0:
        for clause in where:
            # Adjust spacing between existing parsed clause and new clause section
            if len(full_clause) > 0:
                full_clause = f"{full_clause} "

            # Parse the individual clause
            if len(clause) == 2:
                full_clause = f"{full_clause}{clause[0]} = {_wrap_string_values(clause[-1])}"
            elif len(clause) == 3:
                if clause[1] not in LiQLSyntax.comparison_operators:
                    raise errors.exceptions.InvalidOperatorError()
                full_clause = f"{full_clause}{clause[0]} {clause[1]} {_wrap_string_values(clause[-1])}"
            elif type(clause) == str:
                full_clause = f"{full_clause}{clause}"

            # Add the logic statement to join multiple clauses as necessary
            if num_clauses > 1:
                multi_clause_logic = join_logic[0]
                join_logic.remove(join_logic[0])
                full_clause = f"{full_clause} {multi_clause_logic}"

            # Decrement the number of clauses
            num_clauses -= 1

    # Return the fully parsed WHERE clause
    return full_clause


class LiQLSyntax:
    """This class defines lists of syntax elements for use in LiQL queries."""
    comparison_operators = ['=', '!=', '>', '<', '>=', '<=']
    logic_operators = ['AND', 'OR', 'IN', 'MATCHES']
    order_operators = ['ASC', 'DESC']
    container_types = [tuple, set, dict, list]



# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests.test_settings
:Synopsis:       This module is used by pytest to verify the :py:mod:`khoros.objects.settings` functionality.
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  09 Jun 2022
"""

import os
import sys

import pytest

from . import resources
from ...errors import exceptions

# Define a global variable to define when the package path has been set
package_path_defined = False


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionadded:: 4.1.0

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    global package_path_defined
    if not package_path_defined:
        sys.path.insert(0, os.path.abspath('../..'))
        package_path_defined = True


def test_node_setting_retrieval():
    """This function tests the retrieval of API v1 and v2 node settings.

    .. versionadded:: 4.1.0
    """
    if not resources.local_test_config_exists() or not resources.local_helper_exists():
        pytest.skip("skipping local-only tests")

    # Instantiate the Khoros object
    set_package_path()
    khoros_object = resources.instantiate_with_local_helper()

    # Define the elements to query
    test_config = resources.get_testing_config()
    v1_setting = test_config.get('test_data').get('settings').get('node_v1_setting')
    v2_setting = test_config.get('test_data').get('settings').get('node_v2_setting')
    node_id = test_config.get('test_data').get('settings').get('node_id')
    node_type = test_config.get('test_data').get('settings').get('node_type')

    # Retrieve and verify the v1 setting
    setting = khoros_object.settings.get_node_setting(v1_setting, node_id, node_type, convert_json=False)
    assert isinstance(setting, str) or setting is None          # nosec

    # Retrieve and verify the v2 setting
    setting = khoros_object.settings.get_node_setting(v2_setting, node_id, node_type, convert_json=True)
    assert isinstance(setting, dict) or setting is None         # nosec


def test_invalid_node_type_exception():
    """This function tests to confirm that invalid nodes will raise the
    :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError` exception.

    .. versionadded:: 4.1.0
    """
    # Instantiate the Khoros object
    set_package_path()
    khoros_object = resources.initialize_khoros_object()

    # Test most common invalid node types
    for invalid_type in ['boards', 'categories', 'group hubs', 'grouphubs']:
        with pytest.raises(exceptions.InvalidNodeTypeError):
            khoros_object.settings.get_node_setting('custom.pretend_setting', 'fake-node', invalid_type)

    # Test that converting a node to plural also results in the exception
    node_type = resources.get_structure_collection('board')
    with pytest.raises(exceptions.InvalidNodeTypeError):
        khoros_object.settings.get_node_setting('custom.pretend_setting', 'fake-node', node_type)


def test_sso_status_retrieval():
    """This function tests to confirm that the :py:meth:`khoros.core.Khoros.Community.sso_enabled` method
       is able to return a Boolean value.

       .. versionadded:: 5.0.0
    """
    if not resources.local_test_config_exists() or not resources.local_helper_exists():
        pytest.skip("skipping local-only tests")

    # Instantiate the Khoros object
    set_package_path()
    khoros_object = resources.instantiate_with_local_helper()

    # Test the method
    sso_enabled = khoros_object.communities.sso_enabled()
    assert sso_enabled is not None



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.test_ssl_verify
:Synopsis:          This module is used by pytest to test the ability to disable SSL verification on API requests
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     09 Jun 2022
"""

from . import resources


def test_default_core_object_setting():
    """This function tests to verify the ``ssl_verify`` setting is ``True`` by default.

    .. versionadded:: 4.3.0

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    khoros = resources.initialize_khoros_object()
    assert khoros.core_settings.get('ssl_verify') is True       # nosec


def test_core_object_with_param_setting():
    """This function tests to verify the ``ssl_verify`` setting is honored when explicitly defined.

    .. versionadded:: 4.3.0
    """
    defined_setting = {'ssl_verify': False}
    khoros = resources.initialize_khoros_object(use_defined_settings=True, defined_settings=defined_setting,
                                                append_to_default=True)
    assert khoros.core_settings.get('ssl_verify') is False      # nosec


def test_api_global_variable_assignment():
    """This function tests to verify that the ``ssl_verify_disabled`` global variable gets defined appropriately.

    .. versionadded:: 4.3.0
    """
    defined_setting = {'ssl_verify': False}
    khoros = resources.initialize_khoros_object(use_defined_settings=True, defined_settings=defined_setting,
                                                append_to_default=True)
    assert api.ssl_verify_disabled is True


def test_api_should_verify_function():
    """This function tests to verify that the :py:func:`khoros.api.should_verify_tls` function works properly.

    .. versionadded:: 4.3.0
    """
    defined_setting = {'ssl_verify': False}
    khoros = resources.initialize_khoros_object(use_defined_settings=True, defined_settings=defined_setting,
                                                append_to_default=True)
    assert api.should_verify_tls(khoros) is False
    assert api.should_verify_tls() is False


# Import modules and initialize the core object
api, exceptions = resources.import_modules('khoros.api', 'khoros.errors.exceptions')



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.settings
:Synopsis:          This module contains functions specific to settings for various structures and objects
:Usage:             ``from khoros.objects import settings``
:Example:           ``value = settings.get_node_settings(khoros_object, 'custom.purpose', 'my-board')``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     20 May 2021
"""

import json

from .. import api, liql, errors
from ..utils import log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)


def get_node_setting(khoros_object, setting_name, node_id, node_type='board', v1=None, convert_json=False):
    """This function retrieves the value of a specific node setting.

    .. versionchanged:: 4.0.0
       The node type validation has been moved form this function to the lower-level private functions.

    .. versionchanged:: 3.3.2
       The ``convert_json`` parameter has been introduced which optionally converts a JSON string into a dictionary.

    .. versionadded:: 3.2.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param setting_name: The name of the setting field for which to retrieve the value
    :type setting_name: str
    :param node_id: The ID of the node associated with the setting to retrieve
    :type node_id: str
    :param node_type: Defines the node as a ``board`` (default), ``category`` or ``grouphub``
    :type node_type: str
    :param v1: Optionally defines a specific Community API version to use when retrieving the value
    :type v1: bool, None
    :param convert_json: Optionally converts a JSON string into a Python dictionary (``False``  by default)
    :type convert_json: bool
    :returns: The value of the setting for the node
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.GETRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.LiQLParseError`
    """
    # Determine which API version to utilize
    if v1 is None:
        # Leverage APIv2 if the setting name follows the standard v2 naming convention for custom metadata fields
        v1 = False if setting_name.startswith("c_") else True

    # Perform the appropriate API request based on required version
    if v1:
        setting_value = _get_v1_node_setting(khoros_object, setting_name, node_id, node_type)
    else:
        setting_value = _get_v2_node_setting(khoros_object, setting_name, node_id, node_type)

    # Convert the setting value to a dictionary when requested and applicable
    if convert_json and isinstance(setting_value, str):
        if setting_value.startswith('{"') and setting_value.endswith('}'):
            try:
                setting_value = json.loads(setting_value)
            except json.decoder.JSONDecodeError:
                logger.error(f"Failed to convert the setting value for '{setting_name}' from a JSON string to a "
                             "dictionary and therefore it will remain a string when returned.")
    return setting_value


def _validate_node_type(_node_type):
    """This function checks to ensure that a valid node type has been provided for viewing or defining node settings.

    .. versionadded:: 3.2.0

    :param _node_type: Defines the node as a ``board`` (default), ``category`` or ``grouphub``
    :type _node_type: str
    :returns: The node type value which may or may not have been updated
    :raises: :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`
    """
    _node_type_map = {
        'board': 'boards',
        'category': 'categories',
        'grouphub': 'grouphubs'
    }
    if _node_type not in _node_type_map.keys() and _node_type not in _node_type_map.values():
        raise errors.exceptions.InvalidNodeTypeError(val=_node_type)
    elif _node_type in _node_type_map:
        _node_type = _node_type_map.get(_node_type)
    return _node_type


def _get_v1_node_setting(_khoros_object, _setting_name, _node_id, _node_type):
    """This function retrieves a node setting value using the Community API v1.

    .. versionchanged:: 4.0.0
       Support has been introduced for group hubs by leveraging the
       :py:func:`khoros.api.get_v1_node_collection` function.

    .. versionchanged:: 3.3.1
       Fixed an issue with the :py:func:`khoros.api.make_v1_request` function call that was resulting
       in :py:exc:`IndexError` exceptions.

    .. versionadded:: 3.2.0

    :param _khoros_object: The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros]
    :param _setting_name: The name of the setting field for which to retrieve the value
    :type _setting_name: str
    :param _node_id: The ID of the node associated with the setting to retrieve
    :type _node_id: str
    :param _node_type: Defines the node as a ``board`` (default), ``category`` or ``grouphub``
    :type _node_type: str
    :returns: The value of the setting in its original format
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    _collection_uri = api.get_v1_node_collection(_node_type)
    _uri = f"/{_collection_uri}/id/{_node_id}/settings/name/{_setting_name}"
    _settings_data = api.make_v1_request(_khoros_object, _uri, request_type='GET')['response']
    if _settings_data.get('status') == 'error':
        raise errors.exceptions.GETRequestError(status_code=_settings_data['error']['code'],
                                                message=_settings_data['error']['message'])
    return _settings_data['value']['$']


def _get_v2_node_setting(_khoros_object, _setting_name, _node_id, _node_type):
    """This function retrieves a node setting value using the Community API v2 and LiQL.

    .. versionchanged:: 4.0.0
       The node type is now validated before constructing the LiQL query.

    .. versionchanged:: 3.3.3
       Error handling has been introduced to avoid an :py:exc:`AttributeError` exception.

    .. versionchanged:: 3.3.1
       Error handling has been introduced to avoid an :py:exc:`KeyError` exception if the setting field
       is not found, and to return a ``None`` value in that situation.

    .. versionadded:: 3.2.0

    :param _khoros_object: The core :py:class:`khoros.Khoros` object
    :type _khoros_object: class[khoros.Khoros]
    :param _setting_name: The name of the setting field for which to retrieve the value
    :type _setting_name: str
    :param _node_id: The ID of the node associated with the setting to retrieve
    :type _node_id: str
    :param _node_type: Defines the node as a ``board`` (default), ``category`` or ``grouphub``
    :type _node_type: str
    :returns: The value of the setting in its original format
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.GETRequestError`
    """
    _node_type = _validate_node_type(_node_type)
    _query = f"SELECT {_setting_name} FROM {_node_type} WHERE id = '{_node_id}'"
    _settings_data = liql.perform_query(_khoros_object, liql_query=_query)
    _returned_items = liql.get_returned_items(_settings_data, only_first=True)
    try:
        _setting_value = _returned_items.get(_setting_name) if _returned_items.get(_setting_name) else None
    except AttributeError:
        _setting_value = None
    return _setting_value


def define_node_setting(khoros_object, setting_name, setting_val, node_id, node_type='board', return_json=True):
    """This function defines a particular setting value for a given node.

    .. versionchanged:: 4.0.0
       The default value for the ``return_json`` parameter is now ``True`` and group hubs are now supported.

    .. versionchanged:: 3.3.2
       The ``return_json`` parameter has been introduced which returns a simple JSON object (as a ``dict``)
       indicating whether or not the operation was successful. (Currently ``False`` by default)

    .. versionchanged:: 3.3.1
       A minor fix was made to the docstring to correct a Sphinx parsing issue. The function itself was not changed.

    .. versionadded:: 3.2.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param setting_name: The name of the setting field for which to retrieve the value
    :type setting_name: str
    :param setting_val: The value of the setting to be defined
    :type setting_val: str
    :param node_id: The ID of the node associated with the setting to retrieve
    :type node_id: str
    :param node_type: Defines the node as a ``board`` (default), ``category`` or ``grouphub``
    :type node_type: str
    :param return_json: Returns a simple JSON dictionary indicating the operation result (``False`` by default)

                        .. caution:: An unsuccessful REST call will result in the raising of the
                                     :py:exc:`khoros.errors.exceptions.PostRequestError` exception if the
                                     ``return_json`` parameter is set to ``False``.

    :type return_json: bool
    :returns: None (or a dictionary if function call includes the ``return_json=True`` kwarg)
    :raises: :py:exc:`ValueError`, :py:exc:`TypeError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`,
             :py:exc:`khoros.errors.exceptions.InvalidNodeTypeError`,
             :py:exc:`khoros.errors.exceptions.PayloadMismatchError`
    """
    collection_uri = api.get_v1_node_collection(node_type)
    uri = f'/{collection_uri}/id/{node_id}/settings/name/{setting_name}/set'
    payload = {'value': str(setting_val)}
    response = api.make_v1_request(khoros_object, uri, payload, 'POST')['response']
    if response.get('status') != "success" and not return_json:
        raise errors.exceptions.POSTRequestError(status_code=response['error']['code'],
                                                 message=response['error']['message'])
    response = None if not return_json else response
    return response



# -*- coding: utf-8 -*-
"""
:Module:         khoros.utils.tests.test_messages
:Synopsis:       This module is used by pytest to verify that messages function properly
:Created By:     Jeff Shurtliff
:Last Modified:  Jeff Shurtliff
:Modified Date:  09 Jun 2022
"""

import pytest

from . import resources


def get_control_data(test_type):
    """This function retrieves control data to use in unit tests.

    :param test_type: Nickname of the test to be performed
    :type test_type: str
    :returns: Payload control data in dictionary format
    """
    control_data = {
        'node': {'data': {'type': 'message', 'board': {'id': 'my-board'}, 'subject': 'This is the subject line'}},
        'node_id': {'data': {'type': 'message', 'board': {'id': 'my-board'}, 'subject': 'This is the subject line'}},
        'node_url': {'data': {'type': 'message', 'board': {'id': 'studio'}, 'subject': 'This is the subject line'}},
        'body': {'data': {'type': 'message', 'board': {'id': 'my-board'},
                          'subject': 'Welcome', 'body': '<h1>Hello!</h1>'}},
        'welcome_tag': {'data': {'type': 'message', 'board': {'id': 'my-board'}, 'subject': 'Welcome',
                                 'tags': [{'type': 'tag', 'text': 'welcome'}]}},
        '12345_tag': {'data': {'type': 'message', 'board': {'id': 'my-board'}, 'subject': 'Welcome',
                               'tags': [{'type': 'tag', 'text': '12345'}]}},
        'hello_world_tags': {'data': {'type': 'message', 'board': {'id': 'my-board'}, 'subject': 'Welcome',
                                      'tags': [{'type': 'tag', 'text': 'hello'}, {'type': 'tag', 'text': 'world'}]}},
        'str_iter_int_tags': {'data': {'type': 'message', 'board': {'id': 'my-board'}, 'subject': 'Welcome',
                                       'tags': [{'type': 'tag', 'text': 'hello'}, {'type': 'tag', 'text': 'world'},
                                                {'type': 'tag', 'text': '12345'}]}},

    }
    return control_data.get(test_type)


def test_construct_only_subject():
    """This function tests to ensure that a :py:exc:`khoros.errors.exceptions.MissingRequiredDataError` exception
    gets raised when only a subject is passed to the :py:func:`khoros.objects.messages.construct_payload` function.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    with pytest.raises(exceptions.MissingRequiredDataError):
        messages.construct_payload('This is the subject line')


def test_construct_with_node():
    """This function tests constructing payload using properly formatted node data.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_control_data('node')
    payload = messages.construct_payload('This is the subject line', node={"id": "my-board"})
    assert payload == control_data      # nosec


def test_construct_with_node_id():
    """This function tests constructing payload using a Node ID.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_control_data('node_id')
    payload = messages.construct_payload('This is the subject line', node_id='my-board')
    assert payload == control_data      # nosec


def test_construct_with_node_url():
    """This function tests constructing payload using a Node URL.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    node_url = 'https://community.khoros.com/t5/Developer-Discussion/bd-p/studio'
    control_data = get_control_data('node_url')
    payload = messages.construct_payload('This is the subject line', node_url=node_url)
    assert payload == control_data      # nosec


def test_construct_with_body():
    """This function tests constructing payload using a message body.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_control_data('body')
    payload = messages.construct_payload('Welcome', node_id='my-board', body='<h1>Hello!</h1>')
    assert payload == control_data      # nosec


def test_construct_with_one_str_tag():
    """This function tests constructing payload using a single tag in string format.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_control_data('welcome_tag')
    payload = messages.construct_payload('Welcome', node_id='my-board', tags='welcome')
    assert payload == control_data      # nosec


def test_construct_with_one_int_tag():
    """This function tests constructing payload using a single tag in integer format.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_control_data('12345_tag')
    payload = messages.construct_payload('Welcome', node_id='my-board', tags=12345)
    assert payload == control_data      # nosec


def test_construct_with_str_iter_int_tags():
    """This function tests constructing payload providing tags in string, list and integer formats.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_control_data('str_iter_int_tags')
    payload = messages.construct_payload('Welcome', node_id='my-board', tags=('hello', ['world'], 12345))
    try:
        assert payload == control_data      # nosec
    except AssertionError:
        assert_tags_present(payload, ['hello', 'world', '12345'])


def test_construct_with_str_iter_int_tags_ignore():
    """This function tests constructing payload providing tags in string, list and integer formats, and with the
    ``ignore_non_string_tags`` argument set to ``True`` as well.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_control_data('hello_world_tags')
    payload = messages.construct_payload('Welcome', node_id='my-board', tags=('hello', ['world'], 12345),
                                         ignore_non_string_tags=True)
    try:
        assert payload == control_data      # nosec
    except AssertionError:
        assert_tags_present(payload, ['hello', 'world'])


def test_construct_with_tag_iterables():
    """This function tests constructing payload providing tags as a list containing two strings.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_control_data('hello_world_tags')
    list_payload = messages.construct_payload('Welcome', node_id='my-board', tags=['hello', 'world'])
    tuple_payload = messages.construct_payload('Welcome', node_id='my-board', tags=('hello', 'world'))
    set_payload = messages.construct_payload('Welcome', node_id='my-board', tags={'hello', 'world'})
    for payload in (list_payload, tuple_payload, set_payload):
        try:
            assert payload == control_data      # nosec
        except AssertionError:
            assert_tags_present(payload, ['hello', 'world'])


def test_payload_validation():
    """This function tests the validation of the message payload to ensure invalid data raises an exception.

    .. versionadded:: 4.3.0

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    # Test null payload
    with pytest.raises(exceptions.InvalidMessagePayloadError):
        messages.validate_message_payload(payload=None)

    # Test conversion to dictionary when JSON string
    payload = '{"data": {"type": "message", "subject": "This is a message subject"}}'
    payload = messages.validate_message_payload(payload)
    assert isinstance(payload, dict)

    # Test incorrect data type
    with pytest.raises(exceptions.InvalidMessagePayloadError):
        payload = [{'type': 'message'}, {'subject': 'This is a message subject'}]
        messages.validate_message_payload(payload)

    # Test payload that is not wrapped in the 'data' field
    with pytest.raises(exceptions.InvalidMessagePayloadError):
        payload = {'type': 'message', 'subject': 'This is a message subject'}
        messages.validate_message_payload(payload)

    # Test payload that is missing the 'type' sub-field
    with pytest.raises(exceptions.InvalidMessagePayloadError):
        payload = {'subject': 'This is a message subject'}
        messages.validate_message_payload(payload)

    # Test payload that has the incorrect 'type' value
    with pytest.raises(exceptions.InvalidMessagePayloadError):
        payload = {'type': 'article', 'subject': 'This is a message subject'}
        messages.validate_message_payload(payload)

    # Test valid payload
    payload = {'data': {'type': 'message', 'subject': 'This is a message subject'}}
    payload = messages.validate_message_payload(payload)
    assert payload.get('data').get('type') == 'message'


def assert_tags_present(payload, tags_to_find):
    """This function asserts that specific tags are found within API payload.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    :param payload: The payload in which to search for tags
    :type payload: dict
    :param tags_to_find: A list or tuple of tags for which to search in the payload
    :type tags_to_find: list, tuple, set
    :returns: None
    :raises: :py:exc:`AssertionError`
    """
    tags_found = []
    for tag_dict in payload['data']['tags']:
        tags_found.append(tag_dict.get('text'))
    for tag in tags_to_find:
        assert tag in tags_found        # nosec


# Import modules and initialize the core object
messages, exceptions = resources.import_modules('khoros.objects.messages', 'khoros.errors.exceptions')
khoros = resources.initialize_khoros_object()



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.labels
:Synopsis:          This module includes functions that handle labels within a Khoros Community environment
:Usage:             ``from khoros.objects import labels``
:Example:           TBD
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     28 Sep 2021
"""

from .. import api, liql, errors
from ..utils import log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.tags
:Synopsis:          This module includes functions that handle tags within a Khoros Community environment
:Usage:             ``from khoros.objects import tags``
:Example:           ``tags.add_single_tag_to_message('tutorial', 123)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     23 May 2022
"""

from .. import api, liql, errors
from ..utils import core_utils, log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)

ITERABLE_TYPES = (list, tuple, set)


def structure_single_tag_payload(tag_text):
    """This function structures the payload for a single tag.

    .. versionadded:: 2.8.0

    :param tag_text: The tag to be included in the payload
    :type tag_text: str
    :returns: The payload as a dictionary
    :raises: :py:exc:`khoros.errors.exceptions.InvalidPayloadValueError`
    """
    if not isinstance(tag_text, str):
        raise errors.exceptions.InvalidPayloadValueError("The tag text must be in string format")
    payload = {
        "data": {
            "type": "tag",
            "text": tag_text
        }
    }
    return payload


def add_single_tag_to_message(khoros_object, tag, msg_id, allow_exceptions=False):
    """This function adds a single tag to an existing message.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    .. versionadded:: 2.8.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param tag: The tag value to be added
    :type tag: str
    :param msg_id: The unique identifier for the message
    :type msg_id: str, int
    :param allow_exceptions: Determines if exceptions are permitted to be raised (``False`` by default)
    :type allow_exceptions: bool
    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    payload = structure_single_tag_payload(tag)
    api_url = f"{khoros_object.core['v2_base']}/messages/{msg_id}/tags"
    response = api.post_request_with_retries(api_url, payload, khoros_object=khoros_object)
    if response['status'] != 'success':
        api_error = errors.handlers.get_error_from_json(response)
        if allow_exceptions:
            raise errors.exceptions.POSTRequestError(api_error)
        else:
            errors.handlers.eprint(api_error)


def get_tags_for_message(khoros_object, msg_id):
    """This function retrieves the tags for a given message.

    .. versionadded:: 2.8.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param msg_id: The Message ID for the message from which to retrieve tags
    :type msg_id: str, int
    :returns: A list of tags associated with the message
    """
    tag_list = []
    query = f"SELECT text FROM tags WHERE messages.id = '{msg_id}'"     # nosec
    response = liql.perform_query(khoros_object, liql_query=query, verify_success=True)
    entries = api.get_items_list(response)
    for entry in entries:
        tag_list.append(entry['text'])
    return tag_list


def structure_tags_for_message(*tags, khoros_object=None, msg_id=None, overwrite=False, ignore_non_strings=False,
                               wrap_json=False):
    """This function structures tags to use within the payload for creating or updating a message.

    .. versionchanged:: 4.3.0
       Introduced the ``wrap_json`` parameter to wrap the tags in a dictionary within the ``items`` key.

    .. versionchanged:: 4.1.0
       The missing type declaration for the ``overwrite`` parameter has been added to the docstring.

    .. versionadded:: 2.8.0

    :param tags: One or more tags or list of tags to be structured
    :type tags: str, list, tuple, set
    :param khoros_object: The core :py:class:`khoros.Khoros` object

                          .. note:: The core object is only necessary when providing a Message ID as it will be
                                    needed to retrieve the existing tags from the message.

    :type khoros_object: class[khoros.Khoros], None
    :param msg_id: Message ID of an existing message so that its existing tags can be retrieved (optional)
    :type msg_id: str, int, None
    :param overwrite: Determines if tags should overwrite any existing tags (where applicable) or if the tags
                      should be appended to the existing tags (default)
    :type overwrite: bool
    :param ignore_non_strings: Determines if non-strings (excluding iterables) should be ignored rather than
                               converted to strings (``False`` by default)
    :type ignore_non_strings: bool
    :param wrap_json: Determines if the list of tags should be wrapped in the ``{"items": []}`` JSON structure
                      -- In other words, a dictionary rather than a list (``False`` by default)
    :type wrap_json: bool
    :returns: A list of properly formatted tags to act as the value for the ``tags`` field in the message payload
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    formatted_list, existing_tags = [], []
    if msg_id and not any((overwrite, khoros_object)):
        raise errors.exceptions.MissingRequiredDataError('The core Khoros object must be provided when supplying '
                                                         'a Message ID')
    elif msg_id and not overwrite:
        errors.handlers.verify_core_object_present(khoros_object)
        existing_tags = get_tags_for_message(khoros_object, msg_id)
    formatted_list = _format_tag_data(tags, ignore_non_strings)
    existing_tags = _format_tag_data(existing_tags, ignore_non_strings)
    complete_tags = core_utils.merge_and_dedup(formatted_list, existing_tags)
    complete_tags = {'items': complete_tags} if wrap_json else complete_tags
    return complete_tags


def _format_tag_data(_collection, _ignore_non_strings):
    """This function is used by :py:func:`khoros.objects.tags.structure_tags_for_message` to format the tag data.

    .. versionadded:: 2.8.0
    """
    _formatted_list = []
    _collection = _get_low_level_tags(_collection)
    for _tag_or_list in _collection:
        if not isinstance(_tag_or_list, str) and _ignore_non_strings:
            continue
        else:
            _tag_or_list = (str(_tag_or_list),)
        for _tag in _tag_or_list:
            _tag = {
                "type": "tag",
                "text": _tag
            }
            _formatted_list.append(_tag)
    return _formatted_list


def _get_low_level_tags(_collection):
    """This function expands a multi-level iterable to get the low-level tags.

    :param _collection: Iterable with one or more levels or a single tag
    :type _collection: list, tuple, set, str, int, float
    :returns: A list of all low-level tags (pre-formatted)
    """
    _low_level_tags, _iterables = [], []
    for _item in _collection:
        if type(_item) not in ITERABLE_TYPES:
            _low_level_tags.append(_item)
        else:
            _iterables.append(_item)
    while len(_iterables) > 0:
        _item = _iterables.pop()
        for _sub_item in _item:
            if type(_sub_item) not in ITERABLE_TYPES:
                _low_level_tags.append(_sub_item)
            else:
                _iterables.append(_sub_item)
    return _low_level_tags


def add_tags_to_message(khoros_object, tags, msg_id, allow_exceptions=False):
    """This function adds one or more tags to an existing message.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    .. versionadded:: 2.8.0

    ..caution:: This function is not the most effective way to add multiple tags to a message. It is recommended
                that the :py:func:`khoros.objects.messages.update` function be used instead with its ``tags``
                argument, which is more efficient and performance-conscious.

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param tags: One or more tags to be added to the message
    :type tags: str, tuple, list, set
    :param msg_id: The unique identifier for the message
    :type msg_id: str, int
    :param allow_exceptions: Determines if exceptions are permitted to be raised (``False`` by default)
    :type allow_exceptions: bool
    :returns: None
    :raises: :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    tags = (tags, ) if isinstance(tags, str) else tags
    for tag in tags:
        add_single_tag_to_message(khoros_object, str(tag), msg_id, allow_exceptions)



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.test_liql
:Synopsis:          This module is used by pytest to verify that LiQL queries can be performed and parsed successfully.
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     09 Jun 2022
"""

import os
import sys

import pytest

from . import resources

# Define global variables to store LiQL query responses
liql_response, liql_items = {}, None

# Define a global variable to define when the package path has been set
package_path_defined = False


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionchanged:: 4.1.0
       This function now leverages a global variable to ensure it only performs the operation once.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    global package_path_defined
    if not package_path_defined:
        sys.path.insert(0, os.path.abspath('../..'))
        package_path_defined = True


def parse_where_clauses():
    """This function runs through several parsing examples to ensure they all complete successfully.

    .. versionchanged:: 3.4.0
       Updated the name of the :py:func:`khoros.liql.parse_where_clause` function in its respective call.
    """
    # Import the liql module
    from khoros import liql

    # Define each test with a name
    examples = {
        '2-Length Tuple': ('id', 5),
        '3-Length Tuple': ('id', '>', 5),
        'Tuple of Tuples': (('id', 5), ('id', 6)),
        'List of Tuples': [('replies.count(*)', '>', 5), ('replies.count(*)', '<', 10)],
        'Logic Test': (('id', 5), ('id', 6), ('username', '!=', 'John')),
        'Set': {('id', 5), ('replies.count(*)', '=', 0)},
        '1-to-1 Dict': {'id': 5, 'replies.count(*)': 0},
        '1-to-2 Dict': {'id': ('>', 5), 'username': ('<', 10)}
    }

    # Perform each test
    for test_case, example in examples.items():
        # Define which logic to use per test
        if test_case == 'Logic Test':
            logic = ('AND', 'OR')
        else:
            logic = 'AND'

        # Convert any sets to lists for printing below without changing the original identifier
        if type(example) == set:
            print_example = list(example)
        else:
            print_example = example

        # Print results of each test
        print(f"Test Case:\t{test_case}")
        try:
            print(f"Length:\t\t{len(print_example)}")
            print(f"Length[0]:\t{len(print_example[0])}")
        except (KeyError, TypeError):
            pass
        print(f"Example:\t{print_example}")
        where = liql.parse_where_clause(example, logic)
        print(f"Parsed:\t\t{where}\n")
    return True


def perform_test_query(return_items=False):
    """This function performs a LiQL query and saves the response in a global variable.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    :param return_items: Determines if the response should be scoped to only the returned items (``False`` by default)
    :type return_items: bool
    :returns: None
    """
    global liql_response, liql_items
    khoros_object = resources.instantiate_with_local_helper()
    query = "SELECT login FROM users WHERE id = '3'"
    if return_items:
        liql_items = khoros_object.query(query, return_items=return_items)
    else:
        liql_response = khoros_object.query(query, return_items=return_items)


def test_where_clause_parsing():
    """This function tests to confirm that LiQL WHERE clauses are getting parsed properly without failing."""
    set_package_path()
    assert parse_where_clauses() is True        # nosec


def test_liql_query():
    """This function tests to confirm that a standard LiQL query can be performed successfully.

    .. versionadded:: 4.1.0
    """
    if not resources.local_helper_exists():
        pytest.skip("skipping local-only tests")
    set_package_path()
    if not liql_response:
        perform_test_query(return_items=False)
    assert isinstance(liql_response, dict) and liql_response.get('status') == 'success'         # nosec


def test_return_items_option():
    """This function tests the ``return_items`` argument in the :py:meth:`khoros.core.Khoros.query` method.

    .. versionadded:: 4.1.0
    """
    if not resources.local_helper_exists():
        pytest.skip("skipping local-only tests")
    set_package_path()
    if not liql_items:
        perform_test_query(return_items=True)
    assert isinstance(liql_items, list)         # nosec



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.messages
:Synopsis:          This module includes functions that identify environment variables for the khoros library
:Usage:             ``from khoros.utils import environment``
:Example:           ``environment.update_env_variable_names()``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     23 May 2022
"""

import os
import json
import importlib

from .. import errors
from ..utils.core_utils import get_file_type

ENV_VARIABLE_NAMES = ['KHOROS_URL', 'KHOROS_TENANT_ID', 'KHOROS_DEFAULT_AUTH', 'KHOROS_OAUTH_ID', 'KHOROS_OAUTH_SECRET',
                      'KHOROS_OAUTH_REDIRECT_URL', 'KHOROS_SESSION_USER', 'KHOROS_SESSION_PW', 'KHOROS_PREFER_JSON',
                      'KHOROS_LIQL_PRETTY', 'KHOROS_LIQL_TRACK_LSI', 'KHOROS_LIQL_ALWAYS_OK', 'KHOROS_TRANSLATE_ERRORS']

ENV_SETTINGS_MAPPING = {
    'KHOROS_URL': ('community_url',),
    'KHOROS_TENANT_ID': ('tenant_id',),
    'KHOROS_DEFAULT_AUTH': ('auth_type',),
    'KHOROS_OAUTH_ID': ('oauth2', 'client_id'),
    'KHOROS_OAUTH_SECRET': ('oauth2', 'client_secret'),
    'KHOROS_OAUTH_REDIRECT_URL': ('oauth2', 'redirect_url'),
    'KHOROS_SESSION_USER': ('session_auth', 'username'),
    'KHOROS_SESSION_PW': ('session_auth', 'password'),
    'KHOROS_PREFER_JSON': ('prefer_json',),
    'KHOROS_TRANSLATE_ERRORS': ('translate_errors',),
    # TODO: Add LiQL environment variables
}

# Define global variables for the variable names and mapping that can be overwritten by custom values
env_variable_names = ENV_VARIABLE_NAMES
env_settings_mapping = ENV_SETTINGS_MAPPING


def _env_variable_exists(env_variable):
    """This function checks to see if an environment variable is already defined.

    .. versionadded:: 2.2.0

    :param env_variable: The name of the environment variable for which to check
    :type env_variable: str
    :returns: Boolean value indicating if the environment variable already exists
    """
    found = False
    try:
        found = True if os.environ[env_variable] else False
    except KeyError:
        pass
    return found


def _get_env_variable_value(env_variable):
    """This function returns the value of a given environment variable name.

    .. versionadded:: 2.2.0

    :param env_variable: The name of the environment variable to return
    :type env_variable: str
    :returns: The value of the environment variable or ``None`` if the variable name does not exist
    """
    return os.getenv(env_variable)


def get_env_variables():
    """This function retrieves any defined environment variables associate with the khoros library.

    .. versionadded:: 2.2.0

    :returns: A dictionary with any relevant, defined environment variables
    """
    env_settings = {}
    for var_name in env_variable_names:
        if _env_variable_exists(var_name):
            env_settings[var_name] = _get_env_variable_value(var_name)
    return env_settings


def _update_env_list(_orig_name, _custom_name):
    """This function replaces a value in the ``env_variable_names`` global variable with a custom value.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    .. versionadded:: 2.2.0

    :param _orig_name: The original value to replace
    :type _orig_name: str
    :param _custom_name: The custom value that will replace the original value
    :type _custom_name: str
    :returns: None
    """
    if _orig_name in env_variable_names:
        env_variable_names.remove(_orig_name)
        env_variable_names.append(_custom_name)


def _update_env_mapping(_orig_name, _custom_name):
    """This function replaces a dictionary key in the ``env_variable_mapping`` global variable with a custom name.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    .. versionadded:: 2.2.0

    :param _orig_name: The original value to replace
    :type _orig_name: str
    :param _custom_name: The custom value that will replace the original value
    :type _custom_name: str
    :returns: None
    """
    if _orig_name in env_settings_mapping:
        env_settings_mapping[_custom_name] = env_settings_mapping.pop(_orig_name)


def update_env_variable_names(custom_names):
    """This function updates the original environment variable names with custom names when applicable.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    .. versionadded:: 2.2.0

    :param custom_names: A dictionary (or file path to a YAML or JSON file) that maps the original and custom names
    :type custom_names: dict, str
    :returns: None
    """
    if custom_names:
        if type(custom_names) == str:
            custom_names = _import_custom_names_file(custom_names)
        if type(custom_names) == dict:
            for orig_name, custom_name in custom_names.items():
                _update_env_list(orig_name, custom_name)
                _update_env_mapping(orig_name, custom_name)


# Define function to import a YAML helper file
def _import_custom_names_file(file_path):
    """This function imports a YAML (.yml) helper config file.

    .. versionchanged:: 4.0.0
       The :py:mod:`importlib` module is now utilized to import the ``PyYAML`` package where necessary.

    .. versionadded:: 2.2.0

    :param file_path: The file path to the YAML file
    :type file_path: str
    :returns: The parsed configuration data
    :raises: :py:exc:`FileNotFoundError`, :py:exc:`khoros.errors.exceptions.UnknownFileTypeError`
    """
    file_type = get_file_type(file_path)
    with open(file_path, 'r') as cfg_file:
        if file_type == 'yaml':
            yaml = importlib.import_module('yaml')
            custom_names_cfg = yaml.safe_load(cfg_file)
        elif file_type == 'json':
            custom_names_cfg = json.load(cfg_file)
        else:
            raise errors.exceptions.UnknownFileTypeError(file=file_path)
    return custom_names_cfg



# -*- coding: utf-8 -*-
"""
:Module:            khoros.objects.archives
:Synopsis:          This module includes functions that handle archiving messages.
:Usage:             ``from khoros.objects import archives``
:Example:           ``archives.archive(khoros_obj, '123', suggested_url, return_status=True)``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     05 Aug 2021
"""

import warnings

from .. import api, errors
from . import messages
from ..utils import log_utils

# Initialize the logger for this module
logger = log_utils.initialize_logging(__name__)


def archive(khoros_object, message_id=None, message_url=None, suggested_url=None, archive_entries=None,
            aggregate_results=False, include_raw=False):
    """This function archives one or more messages while providing an optional suggested URL as a placeholder.

    .. versionchanged:: 4.1.0
       Made some minor docstring and code adjustments and also removed the following parameters due to the unique
       response format: ``full_response``, ``return_id``, ``return_url``, ``return_api_url``, ``return_http_code``,
       ``return_status``, ``return_error_messages`` and ``split_errors``

       An issue with the :py:func:`khoros.objects.archives.structure_archive_payload` function call was also resolved.

       The optional ``aggregate_results`` parameter has also been introduced.

    .. versionadded:: 2.7.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param message_id: The message ID for the content to be archived
    :type message_id: str, int, None
    :param message_url: The URL of the message to be archived (as an alternative to the ``message_id`` argument)
    :type message_url: str, None
    :param suggested_url: The full URL to suggest to the user when navigating to the archived message
    :type suggested_url: str, None
    :param archive_entries: A dictionary mapping one or more message IDs with accompanying suggested URLs

                            .. note:: Alternatively, a list, tuple or set of message IDs can be supplied which
                                      will be converted into a dictionary with blank suggested URLs.

    :type archive_entries: dict, list, tuple, set, None
    :param aggregate_results: Aggregates the operation results into an easy-to-parse dictionary (``False`` by default)
    :type aggregate_results: bool
    :param include_raw: Includes the raw API response in the aggregated data dictionary under the ``raw`` key
                        (``False`` by default)

                        .. note:: This parameter is only relevant when the ``aggregate_results`` parameter is ``True``.

    :type include_raw: bool
    :returns: Boolean value indicating a successful outcome (default), the full API response or one or more specific
              fields defined by function arguments
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    if not any((message_id, message_url, archive_entries)):
        raise errors.exceptions.MissingRequiredDataError("The message ID or URL must be provided to archive content.")
    if not message_id and message_url:
        message_id = messages.get_id_from_url(message_url)
    api_url = f"{khoros_object.core.get('v2_base')}/archives/archive"
    payload = structure_archive_payload(message_id, suggested_url, archive_entries=archive_entries)
    response = api.post_request_with_retries(api_url, payload, khoros_object=khoros_object,
                                             content_type='application/json')
    results = api.deliver_v2_results(response, full_response=True)
    return aggregate_results_data(results, include_raw) if aggregate_results else results


def unarchive(khoros_object, message_id=None, message_url=None, new_board_id=None, archive_entries=None,
              aggregate_results=False, include_raw=False):
    """This function unarchives one or more messages and moves them to a given board.

    .. versionchanged:: 4.1.0
       Made some minor docstring and code adjustments and also removed the following parameters due to the unique
       response format: ``full_response``, ``return_id``, ``return_url``, ``return_api_url``, ``return_http_code``,
       ``return_status``, ``return_error_messages`` and ``split_errors``

       The optional ``aggregate_results`` parameter has also been introduced.

    .. versionadded:: 2.7.0

    :param khoros_object: The core :py:class:`khoros.Khoros` object
    :type khoros_object: class[khoros.Khoros]
    :param message_id: The message ID for the content to be archived
    :type message_id: str, int, None
    :param message_url: The URL of the message to be archived (as an alternative to the ``message_id`` argument)
    :type message_url: str, None
    :param new_board_id: The board ID of what will be the new parent board of a message getting unarchived
    :type new_board_id: str, None
    :param archive_entries: A dictionary mapping one or more message IDs with accompanying board IDs

                            .. note:: Alternatively, a list, tuple or set of message IDs can be supplied which
                                      will be converted into a dictionary with blank board IDs.

    :type archive_entries: dict, list, tuple, set, None
    :param aggregate_results: Aggregates the operation results into an easy-to-parse dictionary (``False`` by default)
    :type aggregate_results: bool
    :param include_raw: Includes the raw API response in the aggregated data dictionary under the ``raw`` key
                        (``False`` by default)

                        .. note:: This parameter is only relevant when the ``aggregate_results`` parameter is ``True``.

    :type include_raw: bool
    :returns: Boolean value indicating a successful outcome (default), the full API response or one or more specific
              fields defined by function arguments
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`,
             :py:exc:`khoros.errors.exceptions.APIConnectionError`,
             :py:exc:`khoros.errors.exceptions.POSTRequestError`
    """
    if not any((message_id, message_url, archive_entries)):
        raise errors.exceptions.MissingRequiredDataError("The message ID or URL must be provided to archive content.")
    if not message_id and message_url:
        message_id = messages.get_id_from_url(message_url)
    api_url = f"{khoros_object.core.get('v2_base')}/archives/unarchive"
    payload = structure_archive_payload(message_id, new_board_id=new_board_id,
                                        archive_entries=archive_entries, unarchiving=True)
    response = api.post_request_with_retries(api_url, payload, khoros_object=khoros_object,
                                             content_type='application/json')
    results = api.deliver_v2_results(response, full_response=True)
    return aggregate_results_data(results, include_raw) if aggregate_results else results


def structure_archive_payload(message_id, suggested_url=None, new_board_id=None, archive_entries=None,
                              unarchiving=False):
    """This function structures the payload for an archive-related API call.

    .. versionadded:: 2.7.0

    :param message_id: The message ID for the content to be archived
    :type message_id: str, int, None
    :param suggested_url: The full URL to suggest to the user if the user tries to access the archived message
    :type suggested_url: str, None
    :param new_board_id: The board ID of what will be the new parent board of a message getting unarchived
    :type new_board_id: str, None
    :param archive_entries: A dictionary mapping one or more message IDs with accompanying suggested URLs (archiving)
                            or new board IDs (unarchiving)

                            .. note:: Alternatively, a list, tuple or set of message IDs can be supplied which
                                      will be converted into a dictionary with blank suggested URLs (archiving) or
                                      blank new board IDs (unarchiving).

    :type archive_entries: dict, list, tuple, set, None
    :param unarchiving: Indicates that the payload is for an **unarchive** task (``False`` by default)
    :type unarchiving: bool
    :returns: The payload for the API call as a list of dictionaries containing message IDs and/or suggested URLs
    :raises: :py:exc:`khoros.errors.exceptions.MissingRequiredDataError`
    """
    payload = []
    if message_id:
        if suggested_url:
            payload.append(_format_single_archive_entry(message_id, suggested_url, _unarchiving=unarchiving))
        elif new_board_id:
            payload.append(_format_single_archive_entry(message_id, _new_board_id=new_board_id,
                                                        _unarchiving=unarchiving))
        else:
            payload.append(_format_single_archive_entry(message_id))
    if archive_entries:
        if not _valid_entries_type(archive_entries) and len(payload) == 0:
            raise errors.exceptions.MissingRequiredDataError("An invalid 'archive_dict' value was provided and a"
                                                             "message ID or message URL was not provided. Nothing"
                                                             "to archive.")
        elif not isinstance(archive_entries, dict):
            if any((isinstance(archive_entries, tuple), isinstance(archive_entries, list),
                    isinstance(archive_entries, set))):
                archive_entries = _convert_entries_to_dict(archive_entries)
        for entry_id, entry_url in archive_entries.items():
            payload.append(_format_single_archive_entry(entry_id, entry_url))
    return payload


def aggregate_results_data(results, include_raw=False):
    """This function aggregates the results of an archive/unarchive operation into an easy-to-parse dictionary.

    .. versionchanged:: 4.1.1
       This function can now properly handle the ``ARCHIVED`` status when returned.

    .. versionadded:: 4.1.0

    :param results: The results from an archive or unarchive operation
    :type results: list, dict
    :param include_raw: Includes the raw API response in the aggregated data dictionary under the ``raw`` key
                        (``False`` by default)
    :type include_raw: bool
    :returns: A dictionary with fields for ``status``, ``archived``, ``unarchived``, ``failed`` and ``unknown`` or the
              raw response when the API call completely fails, with the optional raw data when requested
    """
    # Initially define the aggregate data
    aggregate_data = {'status': 'success'}
    archived_values = ['ARCHIVING', 'ARCHIVED']
    archived, unarchived, failed, unknown = [], [], [], 0

    # Return the raw error response if the entire API call failed
    if isinstance(results, dict) and results.get('status') == 'error':
        # TODO: Record a log entry for the failed API call
        aggregate_data.update(results)
    elif isinstance(results, list):
        for message in results:
            if isinstance(message, dict) and message.get('archivalStatus') in archived_values:
                archived.append(f"{message.get('msgUid')}")
            elif isinstance(message, dict) and message.get('unarchivalStatus') == 'UNARCHIVING':
                unarchived.append(f"{message.get('msgUid')}")
            elif isinstance(message, dict) and message.get('msgUid'):
                failed.append(f"{message.get('msgUid')}")
            else:
                # TODO: Record a log entry for the unknown result
                unknown += 1

        # Update the aggregate data with the parsed results and return the dictionary
        aggregate_data['archived'] = archived
        aggregate_data['unarchived'] = unarchived
        aggregate_data['failed'] = failed
        aggregate_data['unknown'] = unknown
        if include_raw:
            aggregate_data['raw'] = results
        return aggregate_data


def _valid_entries_type(_entries):
    """This function checks whether or not the ``archive_entries`` argument is a valid type.

    .. versionadded:: 2.7.0

    :param _entries: The ``archive_entries`` value from the parent function
    :returns: Boolean value indicating whether the value is a ``dict``, ``tuple``, ``list`` or ``set``
    """
    return any((isinstance(_entries, dict), isinstance(_entries, tuple),
                isinstance(_entries, list), isinstance(_entries, set)))


def _convert_entries_to_dict(_entries):
    """This function converts a list, tuple or set of archive entries into a dictionary.

    .. versionadded:: 2.7.0

    .. caution:: This is only permitted when the values

    :param _entries:
    :return:
    """
    _new_dict = {}
    for _entry in _entries:
        if isinstance(_entry, str) and not _entry.isnumeric():
            warnings.warn(f"The entry '{_entry}' is not a valid message ID and will be ignored.", RuntimeWarning)

        else:
            _new_dict[str(_entry)] = ""
    return _new_dict


def _format_single_archive_entry(_message_id, _suggested_url=None, _new_board_id=None, _unarchiving=False):
    """This function formats a single entry to be archived.

    .. versionchanged:: 4.1.0
       The ``messageID`` key was incorrect and has been fixed to be ``messageId`` instead.

    .. versionadded:: 2.7.0

    :param _message_id: The ID of the message to be archived
    :type _message_id: str, int
    :param _suggested_url: The full URL to suggest to the user if the user tries to access the archived message
    :type _suggested_url: str, None
    :param _new_board_id: The board ID of a new parent board for a message getting unarchived
    :type _new_board_id: str, None
    :param _unarchiving: Indicates that the payload is for an **unarchive** task (``False`` by default)
    :type _unarchiving: bool
    :returns: The properly formatted archive entry
    """
    _archive_entry = {"messageId": str(_message_id)}
    if _suggested_url and isinstance(_suggested_url, str) and not _unarchiving:
        _archive_entry['suggestedUrl'] = _suggested_url
    elif (_new_board_id and isinstance(_new_board_id, str)) or \
            (_suggested_url and isinstance(_suggested_url, str) and _unarchiving):
        _archive_entry['boardId'] = _new_board_id
    return _archive_entry



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.resources
:Synopsis:          Frequently used resources for performing unit testing
:Usage:             ``from khoros.utils.tests import resources``
:Example:           ``exceptions = resources.import_exceptions_module()``
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     10 Jun 2022
"""

import os
import sys
import importlib

import yaml

# Define global variable to store the YAML test settings
test_config = {}


def set_package_path():
    """This function adds the high-level khoros directory to the sys.path list.

    .. versionadded:: 2.7.4

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    :returns: None
    """
    sys.path.insert(0, os.path.abspath('../..'))


def import_modules(*modules):
    """This function imports and returns one or more modules to utilize in a unit test.

    .. versionadded:: 2.7.4

    :param modules: One or more module paths (absolute) in string format
    :returns: The imported module(s) as an individual object or a tuple of objects
    """
    imported_modules = []
    for module in modules:
        imported_modules.append(importlib.import_module(module))
    tuple(imported_modules)
    return imported_modules if len(imported_modules) > 1 else imported_modules[0]


def initialize_khoros_object(use_defined_settings=False, defined_settings=None, append_to_default=False):
    """This function imports the :py:class:`khoros.core.Khoros` class and initializes an object.

    .. versionchanged:: 4.3.0
       Added support for utilizing the ``defined_settings`` parameter.

    .. versionadded:: 2.7.4

    :returns: The initialized :py:class:`khoros.core.Khoros` object
    """
    set_package_path()
    default_defined_settings = {
        'community_url': 'https://community.example.com',
        'auto_connect': False,
        'tenant_id': 'example',
        'auth_type': 'session_auth',
        'session_auth': {
            'username': 'testuser',
            'password': 'fakePassword123',
        },
    }
    core_module = importlib.import_module('khoros.core')
    if use_defined_settings:
        settings = default_defined_settings if defined_settings is None else defined_settings
        if defined_settings and append_to_default:
            settings = default_defined_settings
            settings.update(defined_settings)
        instantiated_object = core_module.Khoros(defined_settings=settings)
    else:
        instantiated_object = core_module.Khoros(community_url='https://community.example.com', auto_connect=False,
                                                 tenant_id='example', auth_type='session_auth',
                                                 session_auth={'username': 'testuser', 'password': 'fakePassword123'})
    return instantiated_object


def get_structure_collection(structure_type):
    """This function identifies the API collection for a given structure type.

    .. versionadded:: 4.1.0

    :param structure_type: The structure type for which to return the corresponding collection.
    :return:
    """
    structure_map = {
        'board': 'boards',
        'category': 'categories',
        'community': 'community',
        'grouphub': 'grouphubs',
    }
    if structure_type in structure_map.values():
        collection = structure_type
    else:
        collection = structure_map.get(structure_type)
    return collection


def _get_local_helper_file_name(_production=False):
    """This function defines the file name of the local helper file to be used with unit testing.

    .. versionadded:: 4.1.0

    :param _production: Defines whether the helper file is associated with a Production environment
    :type _production: bool, None
    :returns: The file name for the local helper file
    """
    if _production is None:
        _file_name = 'helper.yml'
    else:
        _file_name = 'prod_helper.yml' if _production else 'stage_helper.yml'
    return _file_name


def local_helper_exists(production=False):
    """This function checks to see if a helper file is present in the ``local/`` directory.

    .. versionadded:: 4.1.0

    :param production: Defines whether the helper file is associated with a Production environment
    :type production: bool, None
    :returns: Boolean value indicating whether the local helper file was found
    """
    file_name = _get_local_helper_file_name(production)
    return os.path.exists(f'local/{file_name}')


def local_test_config_exists():
    """This function checks to see if the *khorostest.yml* file is present in the ``local/`` directory.

    .. versionadded:: 4.1.0

    :returns: Boolean value indicating whether the file was found
    """
    return os.path.exists('local/khorostest.yml')


def parse_testing_config_file():
    """This function parses the ``local/khorostest.yml`` file when present.

    .. versionadded:: 4.1.0

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.

    :returns: None
    """
    global test_config
    if local_test_config_exists():
        with open('local/khorostest.yml', 'r') as file:
            test_config = yaml.safe_load(file)


def get_testing_config():
    """This function returns the test config data from the ``local/khorostest.yml`` file when present.

    .. versionadded:: 4.1.0
    """
    if not test_config and local_test_config_exists():
        parse_testing_config_file()
    return test_config


def instantiate_with_local_helper(production=False):
    """This function instantiates a Khoros object using a local helper file for unit testing.

    .. versionadded:: 4.1.0

    :param production: Defines whether the helper file is associated with a Production environment
    :type production: bool, None
    :returns: The instantiated :py:class:`khoros.core.Khoros` object
    """
    file_name = _get_local_helper_file_name(production)
    if local_helper_exists():
        set_package_path()
        core_module = importlib.import_module('khoros.core')
        return core_module.Khoros(helper=f"local/{file_name}")


def instantiate_with_placeholder():
    """This function instantiates a Khoros object with placeholder data.

    .. versionadded:: 5.0.0

    :returns: The instantiated :py:class:`khoros.core.Khoros` object
    """
    set_package_path()
    core_module = importlib.import_module('khoros.core')
    return core_module.Khoros(placeholder=True)



# -*- coding: utf-8 -*-
"""
:Module:            khoros.utils.tests.test_tags
:Synopsis:          This module is used by pytest to verify that tags function properly
:Created By:        Jeff Shurtliff
:Last Modified:     Jeff Shurtliff
:Modified Date:     09 Jun 2022
"""

from . import resources


def test_single_tag_structure():
    """This function tests the :py:func:`khoros.objects.tags.test_single_tag_structure` function.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = {'data': {'type': 'tag', 'text': 'my tag'}}
    payload = tags.structure_single_tag_payload('my tag')
    assert payload == control_data      # nosec


def get_structure_control_data(test_type):
    """This function retrieves the control data to use in tag structure tests.

    :param test_type: The type of test for which to return control data
    :type test_type: str
    :returns: The control data for the given test type
    """
    control_data = {
        'one_tag': [{'type': 'tag', 'text': 'first tag'}],
        'two_tags': [{'type': 'tag', 'text': 'first tag'}, {'type': 'tag', 'text': 'second tag'}],
        'str_int': [{'type': 'tag', 'text': 'first tag'}, {'type': 'tag', 'text': '12345'}],
    }
    return control_data.get(test_type)


def test_message_structure_one_tag():
    """This function tests the :py:func:`khoros.objects.tags.structure_tags_for_message` function with
    a single tag in string format.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_structure_control_data('one_tag')
    data = tags.structure_tags_for_message('first tag')
    assert data == control_data     # nosec


def test_message_structure_two_tags():
    """This function tests the :py:func:`khoros.objects.tags.structure_tags_for_message` function with
    two tags in string format.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_structure_control_data('two_tags')
    data = tags.structure_tags_for_message('first tag', 'second tag')
    assert data == control_data     # nosec


def test_message_structure_one_string_tag_ignore():
    """This function tests the :py:func:`khoros.objects.tags.structure_tags_for_message` function with
    a single tag in string format and the ``ignore_non_strings`` keyword argument set to ``True``.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_structure_control_data('one_tag')
    data = tags.structure_tags_for_message('first tag', ignore_non_strings=True)
    assert data == control_data     # nosec


def test_message_structure_two_string_tags_ignore():
    """This function tests the :py:func:`khoros.objects.tags.structure_tags_for_message` function with
    two tags in string format and the ``ignore_non_strings`` keyword argument set to ``True``.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_structure_control_data('two_tags')
    data = tags.structure_tags_for_message('first tag', 'second tag', ignore_non_strings=True)
    assert data == control_data     # nosec


def test_message_structure_str_int():
    """This function tests the :py:func:`khoros.objects.tags.structure_tags_for_message` function with
    one tag in string format and another as an integer.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_structure_control_data('str_int')
    data = tags.structure_tags_for_message('first tag', 12345)
    assert data == control_data     # nosec


def test_message_structure_str_int_ignore():
    """This function tests the :py:func:`khoros.objects.tags.structure_tags_for_message` function with
    one tag in string format and another as an integer and with ``ignore_non_strings`` set to ``True``.

    .. versionchanged:: 5.0.0
       Removed the redundant return statement.
    """
    control_data = get_structure_control_data('one_tag')
    data = tags.structure_tags_for_message('first tag', 12345, ignore_non_strings=True)
    assert data == control_data     # nosec


# Import modules
tags, exceptions = resources.import_modules('khoros.objects.tags', 'khoros.errors.exceptions')
